DATABASE_API = 'database_api'
SOCIAL_NETWORK_API = 'social_network'
FOLLOW_API = 'follow'
MARKET_HISTORY_API = 'market_history'
PRIVATE_MESSAGE_API = 'private_message'
ACCOUNT_BY_KEY_API = 'account_by_key'
NETWORK_BROADCAST_API = 'network_broadcast_api'
RAW_BLOCK_API = 'raw_block'
BLOCK_INFO_API = 'block_info'

API_LIST = (
    DATABASE_API,
    SOCIAL_NETWORK_API,
    FOLLOW_API,
    MARKET_HISTORY_API,
    PRIVATE_MESSAGE_API,
    ACCOUNT_BY_KEY_API,
    NETWORK_BROADCAST_API,
    RAW_BLOCK_API,
    BLOCK_INFO_API
)